; Progress:
;	3/15/99 Using a port 61h timer. Cleaned up code. At least I've got
;		the motor turning on and off. :)
;
;	3/17/99 Debugged some stack problems. Got it working, barely (read
;		the first sector).
;
;	3/19/99 Does 5 retries now. Doesn't wait for spin-up; much faster.
;		Seems fairly reliable now. Also cleaned up code a bit.
;		Added the '_load' command - almost ready for loadable
;		modules! It works! (most of the time :-)
;
;	3/20/99 Found the "head 1 doesn't work" bug: an AND where an OR
;		should have been. (When combining the disk # and head #)
;
;	3/21/99 Made the motor on/off routines more general, so you can use
;		up to four drives.
;
;	8/15/99	Started overhauling... conversion to Forth, cleaning up...
;
; Bugs:
;
;
; To do:
;	DISK WRITES!!
;
;	Clean up, organize routines better, more uniform interface.
;	improved error handling
;
;	"sense drive status" command
;
;	*FASTER* (don't wait when it's not necessary. Handle motor in a more
;		efficient way)
;
; References:
;	/comp/doc/disk/FLOPPY*	(detailed specs)
;	/c/retro/floppy/fp*.txt	(OCR scanned text on FDC usage)
;	/comp/doc/disk/drives/FDCDEMO.ASM
;	/c/tunes/oberon/floppy.txt
;	/usr/src/linux/drivers/block/floppy.c (Linux)
;	/sys/block/floppy.c ?? (FreeBSD)
;

;; Ports
fdc_output	equ	0x3F2		; (w)
fdc_status	equ	0x3F4		; (r)
fdc_data	equ	0x3F5		; (r/w)


	SECTION .text

init_floppy:
	sprint "Floppy driver",10
	callproc patch_int,	0x26, irq6_handler
	in	al, 0x21
	and	al, 10111111b
	out	0x21, al
	ret



irq6_handler:
	pchr '%'
	inc byte [fd_done]
	push	eax
	irqack	6
	pop	eax
;	sti
	iretd


fd_irq_wait:
.1	cmp byte [fd_done], 0
	jz	.1
	ret


FDR:		;; ( start count addr -- ior )
; (floppy disk read)
fd_load:
;
; Parameters:
;	Starting sector, S (LBA, 0-based like in Unix)
;	Number of sectors, N
;	Destination address, A
;
; Algorithm (same for load & save):
;	Iterate from sector S to S+N-1
;		Convert to C/H/S
;		Load/save sector
;
	mpush	eax,ebx,ecx,edx,esi,edi

	dpop	edi		;destination address
	dpop	ecx		;count
	dpop	eax		;current sector

.loop
	dpush	eax
	call	CHS_convert
	call	read_sector

	;Copy from FD_DMABUF to destination address
	mpop	eax, edi
	push	ecx
	mov	esi, FD_DMABUF	;Source address
	mov	ecx, 128	;128*4 bytes = 512
	cld
	rep	movsd
	pop	ecx
	inc	eax		;Next sector
	loop	.loop

	mpop	eax,ebx,ecx,edx,esi,edi
	ret
eFDR


CHS_convert:	;; ( sector -- CHS )
; CHS format:
;	byte 0	Sector		LSB
;	byte 1	Head
;	byte 2	Cylinder
;	byte 3	x (reserved)	MSB
;
	mpush	eax,edx
	mov	eax, [ebp]
	xor	edx, edx
	div	dword [disk.sectors]
	mov	[ebp+0], dl		;sector
	xor	edx, edx
	div	dword [disk.heads]
	mov	[ebp+1], dl		;head
	mov	[ebp+2], ax		;cylinder
	mpop	eax,edx
	ret
eCHS_convert



send_command:
;
; Parameters:
;	n	Number of bytes to send
;	data	n bytes
;
	mpush	eax,ecx,edx,esi

	mov	ecx, [esp+5*4]	;# of command bytes
	mov byte [fd_done], 0	;clear IRQ flag

	lea	esi, [esp+6*4]
	cld
.l1	mov	dx, fdc_status	;Wait until ready to receive
.l2	in	al, dx
	test	al, 0x80	;repeat while (bit 7 != 1) 
	jz	.l2

	mov	dx, fdc_data	;Send bytes to data register
	lodsd
	out	dx, al
phex eax
	loop	.l1

	call	fd_irq_wait
cr
	mpop	eax,ecx,edx,esi
	ret



motor_on:
;
; AL = Drive #, 0 to 3 (Turns on the motor & selects the drive)
; Clobbers AX, CL, DX
;
	call	motor_build_bitmask
	or	[motor_bits], ah
	jmp short motor_send_cmd


motor_off:
;
; Same as for motor_on
;
	call	motor_build_bitmask
	xor	[motor_bits], ah


motor_send_cmd:			; Helper routine
	or	al, [motor_bits]
	or	al, 0x0C	;Turn on INT/DMA and RESET bits
	mov	dx, fdc_output
	out	dx, al
	ret


motor_build_bitmask:		; Helper routine
	mov	ah, 0x10
	mov	cl, al
	shl	ah, cl		;Build bitmask
	ret



read_status:	;; ( -- stat4-7 stat0-3 )
;
; Read status bytes after sending a command to the FDC
;
; To do:
;	Generalize this routine to work with other commands besides
;		read_sector. (Parameter: # of bytes to read)
;	Possibly, process the information, check for errors..
;
	mpush	eax,ecx,edx,edi
	sub	ebp, 2*4	;dpush 2 items
	mov	edi, ebp
	cld
	mov	ecx, 7		;7 status regs in FDC
.l1	mov	dx, fdc_status	;Wait until ready
.l2	in	al, dx
	test	al, 0x80	;repeat while (bit 7 != 1) 
	jz	.l2
	mov	dx, fdc_data
	in	al, dx
	stosb
	loop	.l1
	mpop	eax,ecx,edx,edi
	ret



read_sector:	;; ( CHS -- )
;
; Parameters:
;	drive
;	cylinder
;	head
;	sector
;
; To do:
;	Generalize this routine for ALL sector operations: read, write, ...?
;
;	Shouldn't the '.retry' loop begin at the read-sector operation?
;	Also, do we need to initialize the DMA every time through the loop?
;
	mpush	eax,ebx,ecx,edx,esi,edi
	movzx	esi, byte [ebp+0]	;edx=sect
	mov	[.sect], esi
	movzx	esi, byte [ebp+1]	;esi=head
	movzx	edi, byte [ebp+2]	;edi=cyl
	DROP

	mov	al, 0		;Drive #
	call	motor_on
;	DPUSH	1000
;	call	DELAY		;If writing, wait 1 sec for spin-up

	mov	ecx, 5		;# of retries allowed
.retry
pchr '*'
	;seek & wait for interrupt
	push	edi		;cylinder
	mov	eax, 0		;EAX = drive # (2 bits)
	mov	ebx, esi	;EBX = head # (1 bit)
	shl	bl, 2
	or	al, bl		;AL = drive & head byte
	push	eax		;drive/head
	push	dword 0x0F	;code
	push	dword 3		;# of command bytes
	call	send_command
	add	esp, 4*4

	DPUSH	2		;wait for head to settle (25ms)
	call	DELAY

	call	init_dma

	;read sector & wait for interrupt
	push	dword	0xFF	;data length
	push	dword	27	;gap length
	push	dword	18	;end of track
	push	dword	2	;512 bytes/sector
	push	dword	[.sect]	;sector
	push	dword	esi	;head
	push	dword	edi	;cylinder
	mov	eax, 0		;EAX = drive # (2 bits)
	mov	ebx, esi	;EBX = head # (1 bit)
	shl	bl, 2
	or	al, bl		;AL = drive & head byte
	push	eax		;drive/head
	push	dword	0xE6	;code (was 0x66 - not multitrack)
	push	dword	9	;9 command bytes
	call	send_command
	add	esp, 10*4

	;read status
	call	read_status
;phex ebp
;phex [ebp+4]
;phex [ebp]
;cr
	test word [ebp], 0xFFD8
	jz	.end		;success!
	dec	ecx
	or	ecx, ecx
	jz near	.fail
	add	ebp, 2*4
	jmp	.retry

.end	mov	al, 0		;Drive #
	call	motor_off
	mpop	eax,ebx,ecx,edx,esi,edi
	ret
.fail
	sprint "Failed, status="
	phex [ebp+4]
	phex [ebp]
	cr
	jmp short .end

.sect	dd	0
eread_sector


init_dma:
;
; prepare DMA controller for transfer
;
; To do: Take read/write mode & buffer address as parameters
;
	mpush	eax
	cli
	mov	al, 0x46	;Put 8237 in read mode
	out	0x0B, al
	out	0x0C, al
	;Send 24-bit address of buffer
;	xor	al, al		;page = 0 (i.e., in first 64k)
	mov	al, FD_DMABUF >> 16	;Page (high byte of 24-bit address)
	out	0x81, al
	mov	ax, FD_DMABUF	;Send low 16 bits of buffer address
	out	0x04, al
	mov	al, ah
	out	0x04, al
	;Send length of buffer (minus one) = 0x200-1 = 0x1FF
	mov	al, 0xFF	;low byte
	out	0x05, al
	mov	al, 0x01	;high byte
	out	0x05, al
	;Enable channel
	mov	al, 2
	out	0x0A, al
	sti
	mpop	eax
	ret



	SECTION .data
disk:
;.cylinders	dd	80
.heads		dd	2
.sectors	dd	18



	SECTION .bss
motor_bits	resb 1		;Motor on/off bits (high nibble) of output reg
stat		resb 8		;Disk status
fd_done		resb 1
