; install.asm - install Retro on a hard drive
; (Currently, only installs to the first IDE drive)

; * No need for several partitions as in Unix (all storage is one)
; * Boot partition might need to be <1024 cylinders, so we'd need two
;   partitions in that case...? (not necessarily)
; * Install procedure: (minimum: steps 1, 3, 6)
;	1) Choose partition (modify partition table if necessary)
;	2) Format partition (optionally, check for bad sectors)
;	3) Copy boot loader (copy first N sectors from floppy to hd)
;	4) Copy objects
;	5) Verify correct installation
;	6) Set partition type (OS); set bootable flag

%include "boot.mac"
%include "macros.mac"
extern	print, print_num, putc, base
extern	getc
extern	ide_read


	SECTION .text
init:	ret


	global install
install:
	pusha
	mpush	[base]
	mov dword [base], 16
	callproc ide_read, 0, 1, buf
	call	choose_ptn	;eax = partition # (0-3)
	cmp	al, 4		;invalid?
	jae	.abort
	call	sure?		;Are you sure??
	or	dl, dl
	jz	.abort

.end	mpop	[base]
	popa
	ret
.abort	callproc print, msg_abort
	jmp short .end


sure?:
;;
;; Returns DL=1 if the user is "sure" about the installation
;;
	push	eax
	callproc print, msg_sure?
	call	getc
	callproc putc, eax
	cr
	xor	dl, dl
	cmp	al, 'y'
	je	.yes
	cmp	al, 'Y'
	jne	.no
.yes	inc	dl
.no	pop	eax
	ret


choose_ptn:
	call	list_ptns
	callproc print, msg_choose_ptn
	call	getc
	callproc putc, eax
	cr
	movzx	eax, al
	sub	al, '1'
	cmp	al, 4
	jb	.ok
	callproc print, msg_invalid_ptn
.ok	ret


list_ptns:
	callproc print, msg_partitions
	mov	esi, buf+0x1be
	xor	ecx, ecx
.l1	inc	ecx
	test byte [esi], 0x80		;boot flag?
	jz	.boot0
	pchr '*'
	jmp short .boot1
.boot0	space
.boot1
	pnum	ecx			;ptn #
	space
	movzx	eax, byte [esi+4]
	pnum	eax			;type
	pchr	9	;tab
	pnum	[esi+8]			;start (lba)
	pchr	9
	pnum	[esi+12]		;size
	cr

	add	esi, 16
	cmp	ecx, 4
	jb near .l1
	ret


	SECTION .data
msg_partitions	db	"Partitions on disk 0:",10
		db	" # type start   size",10,0
msg_choose_ptn	db	"Partition to install Retro on: ",0
msg_invalid_ptn	db	"Invalid partition number.",10,0
msg_sure?	db	"WARNING: You are about to overwrite a partition.",10
		db	"Really (y/N)? ",0
msg_abort	db	"Installation aborted.",10,0


	SECTION .bss
buf	resb	512
ptn	resd	1	; partition to install on
start	resd	1	; starting sector (lba)
size	resd	1	; size (in sectors)
